# Copyright (C) 2011 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['protect']

import sys
import socket
import datetime
import xmlrpclib

from apache_openid.request import Request
from apache_openid.response import Response
from apache_openid.handlers.openidteams.handler import OpenIDTeamsLoginHandler
from apache_openid.utils import OK, Session, log_error

HTTP_SERVICE_UNAVAILABLE = 503


def protect(apache_request):
    """This is the function that protects a directory.

    It should be used as a PythonAccessHandler. e.g.:

      PythonAccessHandler launchpad_lists_handler::protect

    Options used by this handler:

    PythonOption public-team-url http://xmlrpc.example.com/teams
    PythonOption public-team-method isTeamPublic

    In this case, the launchpad_lists_handler module will contact the XMLRPC
    end-point http://xmlrpc.example.com/teams, calling the method
    isTeamPublic(), passing in the team name.  The module expects a boolean
    answer as to whether the named team is public or not.

    For efficiency, this answer is cached in the session data.  The lifetime of
    cache entries is controlled by the `team-cache-lifetime` option.

    PythonOption team-cache-lifetime 60
    """
    lp_req = LaunchpadListsHandler(apache_request)
    try:
        return lp_req.protect()
    finally:
        lp_req.session.save()

class LaunchpadListsHandler:

    def __init__(self, apache_request):
        self.apache_request = apache_request
        self.options = apache_request.get_options()
        self.session = Session(
            self.apache_request,
            timeout=3600,
            lock=False)
        self.action_path = self.options.get('action-path',
            self.apache_request.hlist.location)
        self.request = Request(self.apache_request, self.action_path,
            self.session)
        self.response = Response(self.request, self.action_path, self.session)

    def protect(self):
        # Allow public teams with no further login required.
        aoi_teams_handler = OpenIDTeamsLoginHandler(self.request,
            self.response, self.options, self.session, self.action_path)

        is_public = self.check_public_teams(
            aoi_teams_handler.authorized_teams)
        if is_public is None:
            # cannot determine if team is public or not
            # return a temporary error (503 - SERVICE UNAVAILABLE)
            return HTTP_SERVICE_UNAVAILABLE
        elif is_public:
            return OK
        # Forward request to apache_openid to check membership of private teams
        else:
            return aoi_teams_handler.protect()

    @property
    def cache_lifetime(self):
        """The team cache lifetime from the config options."""
        cache_lifetime = self.options.get('team-cache-lifetime', 3600)
        try:
            cache_lifetime = int(cache_lifetime)
        except ValueError:
            self.error('Bad team-cache-lifetime: %s', cache_lifetime)
            cache_lifetime = 3600
        self.debug('team-cache-lifetime: %s', cache_lifetime)
        return datetime.timedelta(seconds=cache_lifetime)

    def debug(self, message, *args):
        """Write debugging output to the Apache error log.

        Output only occurs if this line appears in the configuration file:

        PythonOption debug 1

        :param message: The %-style string template.
        :type message: string
        :param args: An optional sequence of values to interpolate into
            the `message` template.
        :type args: A *-args tuple
        """
        if self.options.get('debug', '0') == '1':
            self.error(message, *args)

    def error(self, message, *args):
        """Write output to the Apache error log.

        :param message: The %-style string template.
        :type message: string
        :param args: An optional sequence of values to interpolate into
            the `message` template.
        :type args: A *-args tuple
        """
        log_error(message % args)

    def check_public_teams(self, authorized_teams):
        """Check that all authorized teams are also public teams.

        Uses the session to cache the list of public teams, reducing the load
        on the XMLRPC server.  The cache expires after a set amount of time.

        :return: True if all the authorized teams are public.
        :rtype: boolean
        """
        # BarryWarsaw 12-Dec-2008 this could be rather inefficient, but
        # currently Launchpad's XMLRPC only allows us to check one team at a
        # time, and for our primary use case, we only need to check one team,
        # so practically speaking right now it makes no difference.
        cache = self.session.setdefault('public_team_cache', {})
        # The cache maps public team names to the date we should evict the
        # team name from the cache.  Start by seeing if any of the authorized
        # teams are in the cache, and if the cache is still valid.
        authorized_teams = set(authorized_teams)
        query_teams = set()
        now = datetime.datetime.now()
        for team in authorized_teams:
            expiry_date = cache.get(team)
            if expiry_date is None:
                # There's no cache entry for this team, so we'll have to query
                # for it.
                query_teams.add(team)
            elif expiry_date < now:
                # The cache entry for this team has expired.  Evict the cache
                # and query for the team.
                self.debug('Invalidating cache for team: %s', team)
                del cache[team]
                query_teams.add(team)
            else:
                # This team is public.
                self.debug('Cache says, team is public: %s', team)
                pass
        if len(query_teams) == 0:
            # There are no non-public teams to query about.  Short-circuit.
            return True
        # See if there's an XMLRPC server we can use to query about the other
        # teams we couldn't trust our cache about.
        public_team_url = self.options.get('public-team-url')
        if public_team_url is None:
            self.debug('No public team url')
            return False
        public_team_method = self.options.get('public-team-method')
        if public_team_method is None:
            self.debug('No public team method')
            return False
        server = xmlrpclib.ServerProxy(public_team_url)
        method = getattr(server, public_team_method, None)
        if method is None:
            self.error('No method %s() on server: %s',
                       public_team_method, public_team_url)
            return False
        # The cache lifetime is given in seconds, defaulting to 1 hour
        expiration = datetime.datetime.now() + self.cache_lifetime
        # For teams which we couldn't trust our cache about, query the XMLRPC
        # server to see about their visibility.
        for team in query_teams:
            try:
                is_public = method(team)
            except (xmlrpclib.ProtocolError, xmlrpclib.Fault,
                    socket.error), e:
                self.error("XMLRPC Error: %s" % e)
                return None
            self.debug('%s is public? %s', team, is_public)
            if not is_public:
                return False
            cache[team] = expiration
        # If we got here, all teams are public.
        return True

